structure ce3DCamera : CE3DCAMERA =
  struct
    structure rect : CERECT = ceRect;
    structure ang  : CE3DANGLE = ce3DAngle;
    structure vert : CE3DVERTEX = ce3DVertex;
    structure edg  : CE3DEDGE = ce3DEdge;
    structure vec : CE3DVECTOR = ce3DVector;
    structure matr : CE3DMATRIX = ce3DMatrix;
    structure amatr : CE3DADVMATRIX = ce3DAdvMatrix;
    structure poly : CE3DPOLY = ce3DPoly;
    structure Geom : GEOMETRY = Geometry;
    structure math : MATH = Math;

    datatype projectionType = parallel | perspective;

    type camera = vert.vertex*            (* eye point *)
                  vert.vertex*            (* look-at point *)
                  vert.vertex*            (* up point (up vector = eye point -> up point) *)
                  projectionType*         (* type of camera *)
                  real* real* ang.angle*  (* filmback x, y, field of view *)
                  real*                   (* Distance to center of interest from eye point *)
                  matr.t option*          (* transformation matrix for display, if calculated *)
                  matr.t option;          (* transformation matrix for pick/correlation, if calculated *)

    fun experimentalCamera () =
      ((0.0, 50.0, 0.0), (0.0, 0.0, 0.0), (0.0, 50.0, ~10.0),
       perspective, 18.0, 13.2, ang.fromRealDeg(33.398489), 50.0, NONE, NONE)

    fun defaultCamera () =
          let
            val eye = (0.0, 0.0, 50.0);
            val lookAt = (0.0, 0.0, 0.0);
            val distToCenterOfInterest = vec.vectorLength(vec.fromVertices(eye, lookAt));
          in
            (eye, lookAt, (0.0, 10.0, 50.0),
             perspective, 18.0, 13.2, ang.fromRealDeg(33.398489),
             distToCenterOfInterest, NONE, NONE)
          end

    fun cameraToVRP_VPN_VUP ((eye, lap, up,
                             _, backX, _,
                             fieldOfView, _, _, _) : camera) =
          let
            val VPN = vec.fromVertices(lap, eye);
                (* VNP is opposite direction of eye->lap *)
            val VUP = vec.fromVertices(eye, up);
            val focalLength = (backX / 2.0) /
                              math.tan(fieldOfView / 2.0);
            val VRP = vec.subtract(eye : vec.t,
                                   vec.scalarMultiply(vec.normalize(VPN),
                                                      focalLength));
          in
            (VRP, VPN, VUP)
          end

    fun cameraToVMMInput ((_, _, _, camType, backX, backY,
                          fieldOfView, _, _, _) : camera) =
          let
            val focalLength = (backX / 2.0) / math.tan(fieldOfView / 2.0);
            val PRP = (0.0, 0.0, focalLength) : vert.vertex;
            val screenWindow = ((~backX / 2.0, ~backY / 2.0), (backX, backY));

            val frontClipPlane = focalLength - (focalLength / 40.0);
                  (* Input from where? Camera? *)
            val backClipPlane = ~1000.0;
                  (* Input from where? Camera? *)
          in
            (screenWindow, camType, (PRP, frontClipPlane, backClipPlane))
          end

    (* FIXME: I know that the dolly, tumble and track methods
       looks like crap, but uhm, I didn't come up with better ideas yet... *)
    fun dolly ((ep, lap, up, camType, backX, backY, angle, dCOI, matr1, matr2) : camera, dist : real) =
      let
        val direction = vec.fromVertices(ep, lap);
        val normalDir = vec.normalize(direction);
        val (mvX, mvY, mvZ) = vec.scalarMultiply(normalDir, dist);

        val (epX, epY, epZ) = ep;
        val (lapX, lapY, lapZ) = lap;
        val (upX, upY, upZ) = up;
      in
        ((epX + mvX, epY + mvY, epZ + mvZ),
         (lapX+ mvX, lapY+ mvY, lapZ+ mvZ),
         (upX + mvX, upY + mvY, upZ + mvZ),
         camType, backX, backY, angle, dCOI - dist, NONE, NONE)
      end

    fun tumbleX ((ep, lookAt, up, camType, backX, backY, viewAngle, dCOI, matr1, matr2) : camera, rotAngle : ang.angle) =
      let
        val vecZ = vec.fromVertices(lookAt, ep);(* Generate camera coorsys Z-axis *)
        val coiVector = vec.scalarMultiply(vec.normalize(vec.negate(vecZ)), dCOI);
        val coi = vec.add(ep : vec.t, coiVector) : vert.vertex;
        val vecUP = vec.fromVertices(ep, up);(* Generate camera coordsys UP-vector *)
        val matrix = amatr.toVectorSysMatrix(coi, vecZ, vecUP);(* map to camera coordsys *)
        val rotMatrix = matr.rotateX(matrix, rotAngle);(* rotate around X-axis (in camera coordsys) *)
        val matrix = amatr.fromVectorSysMatrix(coi, vecZ, vecUP);(* map back to world coordsys *)
        val matrix = matr.multiply(rotMatrix, matrix);(* map back to world coordsys *)
        val transFun = matr.transformVertexFunction matrix;(* Get complete transformation function *)
      in
        (transFun ep, transFun lookAt, transFun up,
         camType, backX, backY, viewAngle, dCOI,
         NONE, NONE)
      end

    fun tumbleY ((ep, lookAt, up, camType, backX, backY, viewAngle, dCOI, matr1, matr2) : camera, rotAngle : ang.angle) =
      let
        val vecZ = vec.fromVertices(lookAt, ep);(* Generate camera coorsys Z-axis *)
        val coiVector = vec.scalarMultiply(vec.normalize(vec.negate(vecZ)), dCOI);
        val coi = vec.add(ep : vec.t, coiVector) : vert.vertex;
        val vecUP = vec.fromVertices(ep, up);(* Generate camera coordsys UP-vector *)
        val matrix = amatr.toVectorSysMatrix(coi, vecZ, vecUP);(* map to camera coordsys *)
        val rotMatrix = matr.rotateY(matrix, rotAngle);(* rotate around Y-axis (in camera coordsys) *)
        val matrix = amatr.fromVectorSysMatrix(coi, vecZ, vecUP);(* map back to world coordsys *)
        val matrix = matr.multiply(rotMatrix, matrix);(* map back to world coordsys *)
        val transFun = matr.transformVertexFunction matrix;(* Get complete transformation function *)
      in
        (transFun ep, transFun lookAt, transFun up,
         camType, backX, backY, viewAngle, dCOI,
         NONE, NONE)
      end

    fun trackX ((ep, lookAt, up, camType, backX, backY, viewAngle, dCOI, matr1, matr2) : camera, dist : real) =
      let
        val vecZ = vec.fromVertices(lookAt, ep);(* Generate camera coorsys Z-axis *)
        val vecUP = vec.fromVertices(ep, up);(* Generate camera coordsys UP-vector *)
        val matrix = amatr.toVectorSysMatrix(lookAt, vecZ, vecUP);(* map to camera coordsys *)
        val rotMatrix = matr.translateXYZ(matrix, ~dist, 0.0, 0.0);(* move along X axis in camera coordsys *)
        val matrix = amatr.fromVectorSysMatrix(lookAt, vecZ, vecUP);(* map back to world coordsys *)
        val matrix = matr.multiply(rotMatrix, matrix);(* map back to world coordsys *)
        val transFun = matr.transformVertexFunction matrix;(* Get complete transformation function *)
      in
        (transFun ep, transFun lookAt, transFun up,
         camType, backX, backY, viewAngle, dCOI,
         NONE, NONE)
      end

    fun trackY ((ep, lookAt, up, camType, backX, backY, viewAngle, dCOI, matr1, matr2) : camera, dist : real) =
      let
        val vecZ = vec.fromVertices(lookAt, ep);(* Generate camera coorsys Z-axis *)
        val vecUP = vec.fromVertices(ep, up);(* Generate camera coordsys UP-vector *)
        val matrix = amatr.toVectorSysMatrix(lookAt, vecZ, vecUP);(* map to camera coordsys *)
        val rotMatrix = matr.translateXYZ(matrix, 0.0, ~dist, 0.0);(* move along Y axis in camera coordsys *)
        val matrix = amatr.fromVectorSysMatrix(lookAt, vecZ, vecUP);(* map back to world coordsys *)
        val matrix = matr.multiply(rotMatrix, matrix);(* map back to world coordsys *)
        val transFun = matr.transformVertexFunction matrix;(* Get complete transformation function *)
      in
        (transFun ep, transFun lookAt, transFun up,
         camType, backX, backY, viewAngle, dCOI,
         NONE, NONE)
      end

    fun viewOrientationMatrix ((vrpX, vrpY, vrpZ) : vert.vertex,
                               VPN : vec.t,
                               VUP : vec.t) =
      let
        val rotMatr   = amatr.viewVectorsRotationMatrix(VPN, VUP);
        val transMatr = matr.translateXYZMatrix(~vrpX, ~vrpY, ~vrpZ);

        val voMatr = matr.multiply(transMatr, rotMatr);
        
        (*val swapMatr  = matr.swapLRCoordsys(voMatr);*)
          (* swapLRCoordsystems: VRC = Left handed, and WRC = Right handed... Here? *)
      in
        voMatr
      end

    fun viewMappingMatrix ((vpX, vpY), (vpW, vpH)) (* Viewplane boundaries *) =
          let
            val cX = vpX + (vpW / 2.0);
            val cY = vpY + (vpH / 2.0);
            val CW = (cX, cY, 0.0); (* Center of window *)

            fun genMapMatrix perspective
                             ((prpX, prpY, prpZ) : vert.vertex, (* Projection Reference Point*)
                              frontPlaneDist, backPlaneDist) =  (* Front and back clipping planes *)
                  let
                    (* DOP = Direction of projection = vector from PRP to CW *)
                    val (dopX, dopY, dopZ) = vec.fromVertices((prpX,prpY,prpZ), CW);
                    val shear = matr.shearXYMatrix(~(dopX / dopZ), ~(dopY / dopZ));
                    val translate =
                          matr.translateXYZMatrix(~prpX, ~prpY, ~prpZ);
                          (*  +average of npc Min/Max *)
                    val shTrans =
                          matr.multiply(translate, shear);
                    val (_, _, vrpMZ) =
                          matr.transformVertexFunction
                            shTrans
                            (0.0, 0.0, 0.0);
                    val vrpMZBack = vrpMZ + backPlaneDist;
                    val scaled =
                          matr.scaleXYZ(shTrans,
                                        (2.0 * vrpMZ) / (vpW * vrpMZBack), (* 2.0 = Coordsys width *)
                                        (2.0 * vrpMZ) / (vpH * vrpMZBack), (* 2.0 = Coordsys hieght *)
                                        ~1.0 / vrpMZBack);(* ~1.0 = Coordsys depth *)
                  in
                    (scaled,
                     (~vrpMZ / (vrpMZBack),
                     ~(vrpMZ + frontPlaneDist) / (vrpMZBack),
                     ~1.0))
                  end
              | genMapMatrix parallel (* FIXME: Not tested! *)
                             ((prpX, prpY, prpZ) : vert.vertex, (* Projection Reference Point*)
                              frontPlaneDist, backPlaneDist) =  (* Front and back clipping planes *)
                  let
                    (* DOP = Direction of projection = vector from PRP to CW *)
                    val (dopX, dopY, dopZ) = vec.fromVertices((prpX,prpY,prpZ), CW);
                    val shear = matr.shearXYMatrix(~(dopX / dopZ), ~(dopY / dopZ));
                    val shTrans =
                          matr.translateXYZ(shear, ~cX, ~cY, ~frontPlaneDist);
                          (*  +average of npc Min/Max *)
                    val scaled =
                          matr.scaleXYZ(shTrans,
                                        2.0 / vpW,(* 2.0 = Coordsys width *)
                                        2.0 / vpH,(* 2.0 = Coordsys height *)
                                        1.0 / (frontPlaneDist - backPlaneDist));
                                        (* 1.0 = Coordsys height *)
                  in
                    (scaled, (~1.0, 0.0, ~1.0))
                  end
          in
            genMapMatrix
          end                      

      fun clipT (denom : real, num : real, tE : real, tL : real) =
        if Real.> ( denom, 0.0) then
          let
            val t = num / denom;
          in
            if (t > tL) then
              (false,tE,tL)
            else if (t > tE) then
              (true,t,tL)
            else
              (true,tE,tL)
          end
        else if (denom < 0.0) then
          let
            val t = num / denom;
          in
            if (t < tE) then
              (false, tE, tL)
            else if (t < tL) then
              (true,tE,t)
            else
              (true, tE, tL)
          end
        else if (num > 0.0) then
          (false, tE, tL)
        else
          (true, tE, tL)


      (* Liang Barsky clipping algorithm *)
      fun clip3DLine (zMin : real) ((x0, y0, z0) : vert.vertex,(x1, y1, z1) : vert.vertex) =
        let
          val (dX, dZ) = (x1-x0, z1-z0);
          val fail = (false,(0.0, 0.0, 0.0),(0.0, 0.0, 0.0));(* FIXME: Should be SOME of (vertex*vertex) *)
          val (clipped, tMin, tMax) = clipT(~dX - dZ, x0+z0, 0.0, 1.0);(* Right side *)
        in
          if clipped then
            let
              val (clipped, tMin, tMax) = clipT(dX - dZ, ~x0+z0, tMin, tMax);(* Left side *)
            in
              if clipped then
                let
                  val dY = y1 - y0;
                  val (clipped, tMin, tMax) = clipT(dY - dZ, ~y0+z0, tMin, tMax); (* Bottom *)
                in
                  if clipped then
                    let
                      val (clipped, tMin, tMax) = clipT(~dY - dZ, y0+z0, tMin, tMax); (* Top *)
                    in
                      if clipped then
                        let
                          val (clipped, tMin, tMax) = clipT(~dZ, z0 - zMin, tMin, tMax); (* Front *)                                                    
                        in
                          if clipped then
                            let
                              val (clipped, tMin, tMax) = clipT(dZ, ~z0 - 1.0, tMin, tMax); (* Back *)
                            in
                              if clipped then
                                (true,
                                 if Real.> (tMin, 0.0) then
                                   (x0 + (tMin * dX), y0 + (tMin * dY), z0 + (tMin * dZ))
                                 else
                                   (x0, y0, z0),
                                 if Real.< (tMax, 1.0) then
                                   (x0 + (tMax * dX), y0 + (tMax * dY), z0 + (tMax * dZ))
                                 else
                                   (x1, y1, z1)
                                 )
                              else
                                fail
                            end
                          else
                            fail
                        end
                      else
                        fail
                    end
                  else
                    fail
                end
              else
                fail
            end
          else
            fail
        end

      fun viewPortToZFP ((x, y), (width, height)) =
            let
              val zoomFactorX = width / 2.0;(* Must be proportional to viewWindow *)
              val zoomFactorY = height / 2.0;
            in
              (zoomFactorX, zoomFactorY,
               Real.round(x + zoomFactorX),
               Real.round(y + zoomFactorY))
            end

      fun project3DLineTo2D perspective viewPort (point0, point1) =
            (* FIXME: I'm not sure if I should use GUI specific (eXene)
               point/line stuff here... *)
            let
              val (zoomFactorX, zoomFactorY, posX , posY) =
                     viewPortToZFP(viewPort);

              fun projectPoint (xC, yC, zC) =
                    if Real.==(zC, 0.0) then
                      raise ceExceptions.generic("Z-coordinate is zero in project3DLineTo2D")
                      (* To avoid division by zero *)
                    else
                      (Geom.PT{x = (posX + Real.round(xC * zoomFactorX / ~zC)),
                               y = (posY - Real.round(yC * zoomFactorY / ~zC))})  
            in
              (Geom.LINE(projectPoint(point0), projectPoint(point1)))
            end
        | project3DLineTo2D parallel viewPort (point0, point1) = (* FIXME: Not tested! *)
            let
              val (zoomFactorX, zoomFactorY, posX , posY) =
                     viewPortToZFP(viewPort);

              fun projectPoint (xC, yC, _) =
                    (Geom.PT{x = (posX + Real.round(xC * zoomFactorX)),
                             y = (posY - Real.round(yC * zoomFactorY))})  
            in
              (Geom.LINE(projectPoint(point0),projectPoint(point1)))
            end
  end
;